/**
 * 创建节点
 * @param {*} val 列表节点值
 */
function createNode(val) {
  this.val = val;
  this.next = null;
  this.previous = null;
}

/**
 * 遍历节点
 * @param {*} node
 * @param {*} callback
 */
function map(node, callback) {
  var currentNode = node;
  while (currentNode) {
    callback && callback(currentNode);
    currentNode = currentNode.next;
  }
}

/**
 * 获取链表长度
 * @param {*} node
 * @returns
 */
function length(node) {
  var len = 0;
  var currentNode = node;
  while (currentNode) {
    len++;
    currentNode = currentNode.next;
  }
  return len;
}

/**
 * 查找节点
 * @param {*} parameter 查找的对比参数可以是函数
 * @param {*} node 链表的节点
 * @param {*} isDown true为向下查找，false为向上查找
 * @returns
 */
function find(node, parameter, isDown = true) {
  var currentNode = node;
  while (currentNode) {
    if (typeof parameter === 'function') {
      if (parameter(currentNode)) {
        return currentNode;
      }
    } else if (currentNode.val === parameter) {
      return currentNode;
    }
    currentNode = currentNode[isDown ? 'next' : 'previous'];
  }
  return 'not find';
}

/**
 * 获取链表最后一个节点
 * @param {*} node
 * @returns
 */
function last(node) {
  var currentNode = node;
  while (currentNode) {
    if (currentNode.next) {
      currentNode = currentNode.next;
    } else {
      return currentNode;
    }
  }
}

/**
 * 获取链表第一个节点
 * @param {*} node
 * @returns
 */
function first(node) {
  var currentNode = node;
  while (currentNode) {
    if (currentNode.previous) {
      currentNode = currentNode.previous;
    } else {
      return currentNode;
    }
  }
}

/**
 * 向后追加节点
 * @param {*} node
 * @param {*} newVal
 */
function push(node, newVal) {
  var el = last(node);
  var newNode = new createNode(newVal);
  newNode.previous = el;
  el.next = newNode;
}

/**
 * 向前追加节点
 * @param {*} node
 * @param {*} newVal
 */
function unshift(node, newVal) {
  var el = first(node);
  var newNode = new createNode(newVal);
  el.previous = newNode;
  newNode.next = el;
}

/**
 * 插入节点
 * @param {*} node
 * @param {*} parameter
 * @param {*} newVal
 */
function insert(node, parameter, newVal) {
  var el = find(node, parameter);
  if (el !== 'not find') {
    var newNode = new createNode(newVal);
    newNode.next = el.next;
    newNode.next.previous = newNode;
    newNode.previous = el;
    el.next = newNode;
  }
}

/**
 * 删除节点
 * @param {*} node
 * @param {*} parameter
 */
function del(node, parameter) {
  var el = find(node, parameter);
  if (el !== 'not find') {
    el.previous.next = el.next;
    el.next.previous = el.previous;
  }
}

/**
 * 修改节点内容
 * @param {*} node
 * @param {*} parameter
 * @param {*} newVal
 */
function setVal(node, parameter, newVal) {
  var el = find(node, parameter);
  if (el !== 'not find') {
    el.val = newVal;
  }
}

/**
 * 交换节点
 * @param {*} node
 * @param {*} parameter1
 * @param {*} parameter2
 */
function exchange(node, parameter1, parameter2) {
  var el1 = find(node, parameter1);
  var el2 = find(node, parameter2);
  if (el1 !== 'not find' && el2 !== 'not find') {
    var val1 = el1.val;
    el1.val = el2.val;
    el2.val = val1;
  }
}

/**
 * 链表排序
 * @param {*} node
 * @param {*} callback
 */
function sort(node, callback) {
  let len = length(node) - 1;
  while (len > 0) {
    let currentNode = first(node);
    let i = 0;
    while (currentNode && i < len) {
      i++;
      var el1 = currentNode;
      var el2 = currentNode.next;
      if (el1 !== null && el2 !== null) {
        if (callback(el1.val, el2.val)) {
          var val1 = el1.val;
          el1.val = el2.val;
          el2.val = val1;
        } else {
          currentNode = el2;
        }
      } else {
        currentNode = null;
      }
    }
    len--;
  }
}

/**
 * 链表函数
 * @param {*} initVal 初始节点值
 */
export function linkedList(initVal) {
  if (initVal === undefined) {
    return null;
  }
  this.data = new createNode(initVal);
  this.map = function (...p) {
    map(this.data, ...p);
  };
  this.length = function () {
    return length(this.data);
  };
  this.find = function (...p) {
    return find(this.data, ...p);
  };
  this.last = function () {
    return last(this.data);
  };
  this.first = function () {
    return first(this.data);
  };
  this.push = function (...p) {
    push(this.data, ...p);
    return this;
  };
  this.unshift = function (...p) {
    unshift(this.data, ...p);
    this.data = first(this.data);
    return this;
  };
  this.insert = function (...p) {
    insert(this.data, ...p);
    return this;
  };
  this.del = function (...p) {
    del(this.data, ...p);
    return this;
  };
  this.setVal = function (...p) {
    setVal(this.data, ...p);
    return this;
  };
  this.exchange = function (...p) {
    exchange(this.data, ...p);
    return this;
  };
  this.sort = function (...p) {
    sort(this.data, ...p);
    this.data = first(this.data);
    return this;
  };
}
